@@ -12,6 +12,7 @@
 #include <iconv.h>
 #include <inttypes.h>
 #include <linux/types.h>
+#include <unistd.h>
 
 #include "query_capacity.h"
 
@@ -20,7 +21,7 @@
 #define STR_BUF_SIZE		257
 
 #define ATTR_SRC_SYSINFO	'S'
-#define ATTR_SRC_OCF		'O'
+#define ATTR_SRC_SYSFS		'F'
 #define ATTR_SRC_HYPFS		'H'
 #define ATTR_SRC_STHYI		'V'
 #define ATTR_SRC_POSTPROC	'P'	// Note: Post-processed attributes can have multiple origins - would be
@@ -40,7 +41,7 @@
 #endif // htobe32
 
 struct qc_handle {
-	void		 *layer;	// holds a copy of the respective *_values struct (see below),
+	void		 *layer;	// holds a copy of the respective *_values struct
 					// and is filled by looking up the offset via the respective *_attrs table
 	struct qc_attr	 *attr_list;
 	int 		  layer_no;
@@ -59,7 +60,7 @@ struct qc_data_src {
 	char *priv;
 };
 
-extern struct qc_data_src sysinfo, ocf, hypfs, sthyi;
+extern struct qc_data_src sysinfo, sysfs, hypfs, sthyi;
 
 /* Utility functions */
 int qc_ebcdic_to_ascii(struct qc_handle *hdl, char *inbuf, size_t insz);
@@ -73,6 +74,7 @@ struct qc_handle *qc_get_cec_handle(struct qc_handle *hdl);
 struct qc_handle *qc_get_lpar_handle(struct qc_handle *hdl);
 struct qc_handle *qc_get_root_handle(struct qc_handle *hdl);
 struct qc_handle *qc_get_prev_handle(struct qc_handle *hdl);
+struct qc_handle *qc_get_top_handle(struct qc_handle *hdl);
 
 /* Debugging-related functions and variables */
 extern long  qc_dbg_level;
@@ -85,8 +87,6 @@ void qc_debug_indent_inc();
 void qc_debug_indent_dec();
 void qc_mark_dump_incomplete(struct qc_handle *hdl, char *missing_component);
 
-/* Build Customization */
-#define CONFIG_DEBUG_TIMESTAMPS
 
 #ifdef CONFIG_DEBUG_TIMESTAMPS
 #define qc_debug(hdl, arg, ...)	if (qc_dbg_level > 0) { \
